'''
These tools are pulled *very* strongly from the great
ingredients in the `colour` package. They've been
modified and simplified here, in order to work better
with the design of rainbowconnection.
'''

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Polygon
from six.moves import reduce

from colour.algebra import LinearInterpolator
from colour.colorimetry import (
    ILLUMINANTS, ILLUMINANTS_SDS, LIGHTNESS_METHODS, LUMINANCE_METHODS,
    SpectralShape, sd_blackbody, sd_ones, sd_to_XYZ,
    wavelength_to_XYZ)
from colour.plotting import (
    ColourSwatch, COLOUR_STYLE_CONSTANTS, XYZ_to_plotting_colourspace, artist,
    filter_passthrough, filter_cmfs, filter_illuminants, override_style,
    render, plot_single_colour_swatch, plot_multi_functions, plot_single_sd)
from colour.utilities import (domain_range_scale, first_item,
                              normalise_maximum, tstack)

from colour import SpectralDistribution

# define a standard set of color-matching functions
CMFs = first_item(filter_cmfs('CIE 1931 2 Degree Standard Observer').values())

def plot_rainbow(axes=None,
                 wavelength=None, flux=None,
                 cmfs='CIE 1931 2 Degree Standard Observer',
                 **kwargs):
    """
    Plot a simple horizontal rainbow,
    based off colour's plot_single_sd.

    Parameters
    ----------
    ax : matplotlib.axes._subplots.AxesSubplot
        The ax into which the rainbow should be drawn.
    cmfs : string
        The color matching function(s?) to use.

    Returns
    -------
    ax : matplotlib.axes._subplots.AxesSubplot
        The ax into which the rainbow was drawn.
    """

    if axes is None:
        axes = plt.gca()

    # pull out the CMFss
    cmfs = first_item(filter_cmfs(cmfs).values())

    if wavelength is None:
        # create a grid of wavelengths (at which CMFss are useful)
        wavelength = cmfs.wavelengths

    ok = ((wavelength >= np.min(cmfs.wavelengths)) &
          (wavelength <= np.max(cmfs.wavelengths)))

    # create colors at theose wavelengths
    colours = XYZ_to_plotting_colourspace(
        wavelength_to_XYZ(wavelength[ok], cmfs),
        ILLUMINANTS['CIE 1931 2 Degree Standard Observer']['E'],
        apply_encoding_cctf=False)

    # normalize the colors to their maximum?
    colours = COLOUR_STYLE_CONSTANTS.colour.colourspace.encoding_cctf(
        normalise_maximum(colours))

    # create y values that will be plotted (these could be spectrum)
    if flux is None:
        flux = np.ones_like(wavelength)
    x_min, x_max = min(wavelength), max(wavelength)
    y_min, y_max = 0, max(flux) + max(flux) * 0.05

    # create a polygon to define the top of the bars?
    polygon = Polygon(
        np.vstack([
            (x_min, 0),
            tstack([wavelength[ok], flux[ok]]),
            (x_max, 0),
        ]),
        facecolor='none',
        edgecolor='none')
    axes.add_patch(polygon)

    # draw bars, with the colors at each vertical stripe
    padding = 0.2
    axes.bar(
        x=wavelength[ok] - padding/2,
        height=max(flux[ok]),
        width=1 + padding,
        color=colours,
        align='edge')

    # plot the actual spectrum?
    # axes.plot(wavelength, values, color=COLOUR_STYLE_CONSTANTS.colour.dark)

    return axes

def rainbow_spectrum(axes=None,
                 wavelength=None, flux=None,
                 cmfs='CIE 1931 2 Degree Standard Observer',
                 rainbowtop=np.inf,
                 **kwargs):
    """
    (This is still *real* blarg-y*.)
    Plot a spectrum, with a rainbow underneath.

    Parameters
    ----------
    ax : matplotlib.axes._subplots.AxesSubplot
        The ax into which the rainbow should be drawn.
    cmfs : string
        The color matching function(s?) to use.

    Returns
    -------
    ax : matplotlib.axes._subplots.AxesSubplot
        The ax into which the rainbow was drawn.
    """

    if axes is None:
        axes = plt.gca()

    if wavelength is None:
        # create a grid of wavelengths (at which CMFss are useful)
        wavelength = CMFs.wavelengths

    # create y values that will be plotted (these could be spectrum)
    if flux is None:
        flux = np.ones_like(wavelength)

    # pull out only the values that *can* be converted to colors
    ok = ((wavelength >= np.min(CMFs.wavelengths)) &
          (wavelength <= np.max(CMFs.wavelengths)))

    w, f = wavelength[ok], flux[ok]

    # clip the top of the box?
    f = np.minimum(f, rainbowtop)

    XYZ = wavelength_to_XYZ(w)

    # create colors at theose wavelengths
    colours = XYZ_to_plotting_colourspace(XYZ)

    # normalize the colors to their maximum?
    #colours = COLOUR_STYLE_CONSTANTS.colour.colourspace.encoding_cctf(
    #    normalise_maximum(colours))
    colours = np.maximum(0, colours/np.max(colours))

    x_min, x_max = min(w), max(w)
    y_min, y_max = 0, max(f) + max(f) * 0.05

    # create a polygon to define the top of the bars?
    polygon = Polygon(
        np.vstack([
            (x_min, 0),
            tstack([w, f]),
            (x_max, 0),
        ]),
        facecolor='none',
        edgecolor='none')
    axes.add_patch(polygon)

    # draw bars, with the colors at each vertical stripe
    padding = 0.0
    dw = np.mean(np.diff(w))
    axes.bar(
        x=w,
        height=f,
        width=dw,
        color=colours,
        clip_path=polygon,
        align='edge',
        clip_on=True)

    return axes



from .atmosphere import *

class HotJupiter(DiscreteAtmosphere):

    def read_transmission(self, TiO=False, **kwargs):
        '''
        Read a Hot Jupiter's transmission spectrum in from a file
        and define the required attributes:

            ._wavelengths (with units of wavelength)
            ._tau_zenith_reference (unitless)
            .H (with units of length)
            .radius (with units of length)
        '''

        # load a transmission spectrum from Fortney et al. (2010)
        if TiO:
            choice = 'lambda_1500K_g10_wTiOVO.dat'
        else:
            choice = 'lambda_1500K_g10_noTiOVO.dat'
        filename = os.path.join(data_directory, f'fortney/{choice}')
        d = ascii.read(filename, comment='#')

        # define the wavelength grid
        self._wavelength = d['wavelength'].data*1e3*u.nm

        # store the effective transit radius
        radiusinkm = d['radiusinkm'].data*u.km
        self._transit_radius = radiusinkm.to('Rjup')

        # pick an (arbitrary-ish) reference radius
        reference_radius = np.min(radiusinkm)

        # define geometry of the atmosphere (how spherical?)
        mu = 2.3
        self.T = 1500*u.K
        g = 10*u.m/u.s**2
        self.H = (con.k_B*self.T/mu/g/con.m_p).to('km')
        self.radius = reference_radius

        # calculate the optical depth at zenith
        z_over_H = ((radiusinkm - reference_radius)/self.H).decompose()
        tau_slant = np.exp(z_over_H)
        self._tau_zenith_reference = (tau_slant/self.fortney_factor()).decompose()



from .atmosphere import *

class Earth(DiscreteAtmosphere):
    '''
    Earth's transmission spectrum.
    '''

    def read_transmission(self, **kwargs):
        '''
        Read Earth's transmission spectrum in from a file
        and define the required attributes:

            ._wavelengths (with units of wavelength)
            ._tau_zenith_reference (unitless)
            .H (with units of length)
            .radius (with units of length)
        '''

        # load the ESO vertical transmission through Earth's atmosphere
        filename = os.path.join(data_directory, 'earthtransmission.txt')
        d = ascii.read(filename, comment='#')


        # calculate the optical depth at zenith
        self._wavelength = d['wavelength'].data*1e3*u.nm
        self._transmission_zenith = d['transmission'].data


        # define geometry of the atmosphere (how spherical?)
        mu = 29
        self.T = 273*u.K
        g = 9.8*u.m/u.s**2
        self.H = (con.k_B*self.T/mu/g/con.m_p).to('km')
        self.radius = 1*u.Rearth

        # the file is for Cerro Paranal at 2640m
        # let's move it down to sea level
        tau_zenith_paranal = -np.log(self._transmission_zenith)
        paranal_altitude = 2640*u.m
        factor = np.exp(paranal_altitude/self.H)
        self._tau_zenith_reference = tau_zenith_paranal*factor
        # to get back to Boulder, set altiude=0.25(=2/8km)



'''Tools for resampling array from grid of independent variables to another.'''

import numpy as np
import scipy.interpolate
import matplotlib.pyplot as plt

def binsizes(x):
	'''If x is an array of bin centers, calculate what their sizes are.
		(assumes outermost bins are same size as their neighbors)'''

	binsize = np.zeros_like(x)
	binsize[0:-1] = (x[1:] - x[0:-1])
	binsize[-1] = binsize[-2]
	return binsize

def plotboxy(x, y, **kwargs):
	'''
	Plot with boxes, to show the left and right edges of a box. This is useful,
	for example, to plot flux associated with pixels, in case you are trying to
	do a sub-pixel resample or interpolation or shift.

	(kwargs are passed on to plt.plot)
	'''

	# what are the edges of the bins (making a guess for those on the ends)
	xbinsize = binsizes(x)
	xleft = x - xbinsize/2.0
	xright = x + xbinsize/2.0

	# create a array that doubles up the y values, and interleaves the edges
	plot_x = np.vstack((xleft,xright)).reshape((-1,),order='F')
	plot_y = np.vstack((y,y)).reshape((-1,),order='F')

	# plot those constructed arrays
	plt.plot(plot_x, plot_y, **kwargs)

def fluxconservingresample(xin_unsorted, yin_unsorted, xout, test=False, visualize=False, demo=False,
							treatnanas=0.0):
	'''
	Starting from some initial x and y, resample onto a different grid
	(either higher or lower resolution), while conserving total flux.

	When including the entire range of xin, sum(yout) == sum(yin) should be true.
	'''

	# sort to make sure x is strictly increasing
	s = np.argsort(xin_unsorted)
	xin = xin_unsorted[s]
	yin = yin_unsorted[s]

	# set up the bins, to calculate cumulative distribution of y?
	xinbinsize = binsizes(xin)
	xinleft = xin - xinbinsize/2.0
	xinright = xin + xinbinsize/2.0

	# the first element should be the left edge of the first pixel
	# last element will be right edge of last pixel
	xinforcdf = np.hstack([xinleft, xinright[-1]])

	# to the left of the first pixel, assume flux is zero
	yinforcdf = np.hstack([0, yin])

	# correct for non-finite
	bad = np.isnan(yinforcdf)
	yinforcdf[bad] = treatnanas

	# calculate the cumulative distribution function of the flux (at pixel edge locations)
	cdfin = np.cumsum(yinforcdf)

	# create an interpolator for that
	cdfinterpolator = scipy.interpolate.interp1d(xinforcdf, cdfin,
						kind='linear',
						bounds_error=False,
						fill_value=(0.0, np.sum(yin)))

	# calculate bin edges (of size len(xout)+1)
	xoutbinsize = binsizes(xout)
	xoutleft = xout - xoutbinsize/2.0
	xoutright = xout + xoutbinsize/2.0
	xoutcdf = np.hstack([xoutleft, xoutright[-1]])

	# interpolate the CDF onto those bin edges
	cdfout = cdfinterpolator(xoutcdf)

	# take the derivative of the CDF to get the flux per resampled bin
	# (xout is the center of the bin, and yout is the flux in that bin)
	yout = np.diff(cdfout)

	if visualize:
		fi, (ax_cdf, ax_pdf) = plt.subplots(2,1, sharex=True, figsize=(9,6))
		inkw = dict(color='black', alpha=1, linewidth=3, marker='.', markeredgecolor='none')
		outkw = dict(color='darkorange', alpha=1, linewidth=1, marker='.', markersize=8, markeredgecolor='none')


		legkw = dict(fontsize=10, frameon=False, loc='upper left', bbox_to_anchor=(1, 1))

		# plot the PDFs
		plt.sca(ax_pdf)
		plt.ylabel('Flux per (Original) Pixel')
		plt.xlabel('Pixel')
		# plot the original pixels (in df/dpixel to compare with resampled)
		plotboxy(xin, yin/xinbinsize,
					label='Original Pixels', **inkw)


		# what would a bad interpolation look like?
		badinterpolation = scipy.interpolate.interp1d(xin, yin/xinbinsize,
							kind='linear', bounds_error=False, fill_value=0.0)
		plt.plot(xout, badinterpolation(xout),
						color='cornflowerblue', alpha=1,
						linewidth=1, marker='.',
						markersize=8, markeredgecolor='none',
						label='Silly Simple Interpolation')

		# plot the flux-conserving resampled data (again, in df/d"pixel")
		plt.plot(xout, yout/xoutbinsize, label='Flux-Conserving Interpolation', **outkw)
		plt.legend(**legkw)

		# plot the CDFs
		plt.sca(ax_cdf)
		plt.ylabel('Cumulative Flux (from left)')

		# plot the original CDF
		plt.plot(xinforcdf, cdfin,
					label='Original Pixels', **inkw)

		# plot the interpolated CDF
		plt.plot(xoutcdf, cdfout,
			label='Flux-Conserved Resample', **outkw)
		#plt.legend(**legkw)
		if demo:
			a = raw_input("Pausing a moment to check on interpolation; press return to continue.")

		plt.tight_layout(rect=[0.0, 0.0, 0.67, 1])
		print('{:>6} = {:.5f}'.format('Actual', np.sum(yin)))
		print('{:>6} = {:.5f}'.format('Silly', np.sum(badinterpolation(xout)*xoutbinsize)))
		print('{:>6} = {:.5f}'.format('CDF', np.sum(yout)))


	# return the resampled y-values
	return yout

def bintogrid(x, y, unc=None, newx=None, dx=None, weighting='inversevariance', drop_nans=True):
	'''
	x = the independent variable (wavelength)
	y = the measurement (transit depth)
	unc = the uncertainty on the measurement (sigmas on the transit depths)
	newx = a (linearly) uniformly spaced grid onto which we should bin
	dx = if newx isn't set, then create a grid with this fixed spacing

	Right now, this works only on 1-dimenions arrays,
	but in the long term we should update it to work in
	for arbitrary N-dimensional arrays.

	Arrays must be ordered from decreasing to increasing x.
	'''

	# make up a grid, if one wasn't specified
	if newx is None:
		newx = np.arange(np.min(x), np.max(x) + dx, dx)

	# don't complain about zero-divisions in here (just make uncertainties -> infinity)
	with np.errstate(divide='ignore',invalid='ignore'):

		# resample the sums onto that new grid (in logarithmic space)
		if unc is None:
			weights = np.ones_like(x)
		else:
			if weighting == 'inversevariance':
				weights = 1/unc**2
		numerator = fluxconservingresample(x, y*weights, newx, visualize=False)
		denominator = fluxconservingresample(x, weights, newx, visualize=False)

		# the binned weighted means on the new grid
		newy = numerator/denominator

		# the standard error on the means, for those bins
		newunc = np.sqrt(1/denominator)

	if drop_nans:
		ok = np.isfinite(newy)
	else:
		ok = np.ones_like(newx).astype(np.bool)
	# return binned x, y, unc
	if unc is None:
		return newx[ok], newy[ok]
	else:
		return newx[ok], newy[ok], newunc[ok]


def bintoR(x, y, unc=None, R=50, xlim=None, weighting='inversevariance'):
	'''
	x = the independent variable (wavelength)
	y = the measurement (transit depth)
	unc = the uncertainty on the measurement (sigmas on the transit depths)
	R = the resolution (lambda/dlambda) to bin everything to
	xlim = the limits of the new binned grid
			[if None, will be created from data]

	Right now, this works only on 1-dimenions arrays,
	but in the long term we should update it to work in
	for arbitrary N-dimensional arrays.

	Arrays must be ordered from decreasing to increasing x.
	'''

	# create a new grid of x at the given resolution
	lnx = np.log(x)
	dnewlnx = 1.0/R


	# set the limits of the new xgrid (in log space)
	if xlim is None:
		# use the input grid to set the limits
		lnxbottom, lnxtop = np.min(lnx), np.max(lnx)
	else:
		# use the custom xlim to set the limits
		lnxbottom, lnxtop = xlim

	# create a new, log-uniform grid of x values
	newlnx = np.arange(lnxbottom, lnxtop + dnewlnx, dnewlnx)

	# now do the binning on a uniform grid of lnx

	if unc is None:
		blnx, by = bintogrid(lnx, y, unc, newx=newlnx, weighting=weighting)
		return np.exp(blnx), by
	else:
		blnx, by, bunc = bintogrid(lnx, y, unc, newx=newlnx, weighting=weighting)
		return np.exp(blnx), by, bunc




def testgrid(dx=1.0, N=500, snr=1000, seed=42, irregular=False):
	'''
	a simple test to show how binning to a grid works
	'''


	print('Testing binning {} {} data binned to dx={}\n'.format(N, {True:'irregularly gridded', False:'gridded'}[irregular], dx))


	# set up fake arrays
	np.random.seed(seed)
	if irregular:
		x = np.sort(np.random.uniform(0.5, 25, N))
	else:
		x = np.linspace(0.5, 25, N)
	unc = np.ones_like(x)/snr
	model = np.zeros_like(x) + np.exp(-0.5*(x-5)**2/2**2)*0.01# x*0.0002
	y = np.random.normal(model, unc, N)


	bx, by, bunc = bintogrid(x, y, unc, dx=dx)

	# plot the demonstrations
	plt.figure()
	plt.errorbar(x, y, unc, color='gray', linewidth=0, elinewidth=1, alpha=0.1)
	plt.errorbar(bx, by, bunc, color='black', linewidth=0, elinewidth=3)
	plt.xlabel('x'); plt.ylabel('y')
	plt.title('{} {} data binned to dx={}'.format(N, {True:'irregularly gridded', False:'gridded'}[irregular], dx))
	plt.show()

	# return the results
	return bx, by, bunc


def testR(R=10, N=500, snr=1000, seed=42, irregular=False):
	'''
	a simple test to show how binning to constant R works
	'''

	print('Testing binning {} {} data binned to R={}\n'.format(N, {True:'irregularly gridded', False:'gridded'}[irregular], R))

	# set up fake arrays
	np.random.seed(seed)
	if irregular:
		x = np.sort(np.random.uniform(0.5, 25, N))
	else:
		x = np.linspace(0.5, 25, N)
	unc = np.ones_like(x)/snr
	model = np.zeros_like(x) + np.exp(-0.5*(x-5)**2/2**2)*0.01# x*0.0002
	y = np.random.normal(model, unc, N)

	bx, by, bunc = bintoR(x, y, unc, R=R)

	# plot the demonstrations
	plt.figure()
	plt.errorbar(x, y, unc, color='gray', linewidth=0, elinewidth=1, alpha=0.1)
	plt.errorbar(bx, by, bunc, color='black', linewidth=0, elinewidth=3)
	plt.xscale('log')
	plt.xlabel('x'); plt.ylabel('y')
	plt.title('{} {} data binned to R={}'.format(N, {True:'irregularly gridded', False:'gridded'}[irregular], R))
	plt.show()

	# return the results
	return bx, by, bunc

def testFCR(supersample=True):
	'''this function tests out the resampling code

			supersample=True
				means that there will be multiple new pixels per original pixel

			supersample=False
				means that there will be fewer new pixels than original pixels
	'''

	print("Testing flux-conserving resampling, when {}supersampling the original grid\n".format({True:'', False:'not '}[supersample]))
	xinitial = np.arange(39,47)
	yinitial = np.random.uniform(0.0, 0.1, len(xinitial))
	if supersample:
		xresample = np.linspace(np.min(xinitial) - 1.5, np.max(xinitial) + 1.5, 50)
	else:
		xresample = np.linspace(np.min(xinitial) - 1.5, np.max(xinitial) + 1.5, 5)
	yresample = fluxconservingresample(xinitial, yinitial, xresample, visualize=True)
	plt.show()

def test():
	testFCR()
	testFCR(supersample=False)
	testR(R=10)
	testR(R=10, irregular=True)
	testgrid(dx=0.5)
	testgrid(dx=0.5, irregular=True)



from ..imports import *
from ..units import determine_quantity
from ..plottingtools import setup_axes_with_rainbow
from .sunset import Sunset

class Atmosphere:
    def __repr__(self):
        '''
        How should this object appear as a string?

        Returns
        -------
        s : str
            A simple string representation.
        '''
        scale = (self.H/self.radius).decompose()
        #H/R={scale},
        return f'{self.__class__.__name__}Atmosphere ({self.zenith_angle} from zenith, {self.altitude} scale heights from reference)'

    # FIXME (maybe both spectrum and atmosphere should inherit from the same thing?)
    def wavelength(self, wavelength=None):
        '''
        A wrapper to ensure at least some grid of wavelengths
        gets defined. A default grid will be assumed, unless
        any wavelength array at all is passed.
        '''

        # make sure at least some wavelengths are defined
        if wavelength is None:
            wavelength = self.default_wavelengths
        return wavelength.to('nm')

    def plot(self,  ax=None,
                    wavelength=None,
                    rainbow=True,
                    color=None,
                    style='dark_background',
                    figsize=(5.0, 2.5),
                    **kwargs):
        '''
        A quick tool to plot the transmission through the atmosphere.

        Parameters
        ----------
        ax : matplotlib.axes._subplots.AxesSubplot
            Specify an axes object into which
            this plot should be drawn. This allows
            overplotting multiple spectra, for example
            with a structure like
                ```
                ax = Thermal(teff=5800*u.K, radius=1*u.Rsun).plot()
                ax = Sun.plot(ax)
                ```
            To overplot on current axes use `ax=plt.gca()`.

        wavelength : astropy.units.quantity.Quantity
            A grid of wavelengths on which the spectrum should
            be plotted. If None, the function defaults to
            covering visible wavelengths at 1nm resolution.

        rainbow : bool
            Should we add an extra rainbow above the plot,
            to indicate how wavelengths match to visible light?

        color : str
            The color for drawing the spectrum.
            'auto' represents the actual visible color.

        '''

        # set up to use a dark background for the plot; make sure units match
        with plt.style.context(style), quantity_support():

            # setup the basic axes
            ax = setup_axes_with_rainbow(ax=ax, rainbow=rainbow, figsize=figsize)

            # make sure at least some wavelengths are defined
            w = self.wavelength(wavelength)

            # pull out the spectrum
            t = self.transmission(w)

            plt.plot(w, t*100, color=color, label=self, **kwargs)

            # add the axis labels
            wunit = w.unit.to_string('latex_inline')
            plt.xlabel(f'Wavelength ({wunit})')
            plt.ylabel(f'Transmission (%)')

            plt.xlim(np.min(w).value, np.max(w).value)

        return ax

    def set_elevation_angle(self, elevation=90*u.deg):
        '''
        Set the elevation angle above the horizon.

        Parameters
        ----------
        elevation : astropy.units.quantity.Quantity
            The angle above the horizon along which
            the transmission of the atmosphere should
            be calculated.
        '''
        self.set_zenith_angle(90*u.deg - elevation)

    def set_zenith_angle(self, zenith_angle=0*u.deg):
        '''
        Set the angle from zenith.

        Parameters
        ----------
        zenith_angle : astropy.units.quantity.Quantity
            The angle away from zenith along which
            the transmission of the atmosphere should
            be calculated.
        '''

        # set the zenith angle and airmass
        self.zenith_angle = np.minimum(zenith_angle, 90*u.deg)
        self.airmass = 1/np.cos(self.zenith_angle)

    def set_altitude(self, altitude=0):
        '''
        Set how many atmospheric scale heights we are
        above or below the reference radius of this
        atmosphere.

        Parameters
        ----------
        altitude : float
            The altitude at which we should be floating
            in the atmosphere, in units of scale heights.
            For example, +1.0 is one scale height above
            the reference radius and therefore a more
            transparent atmosphere, and -1.0 is one scale
            height deeper than the reference radius and
            therefore a more opaque atmosphere.
            Negative altitudes are probably a weird concept
            if the reference radius is the surface of a
            rocky planet, but they're quite reasonable
            on gas-dominated planets.
        '''
        self.altitude = altitude
        self.tau_zenith = self._tau_zenith_reference*np.exp(-altitude)

        # guess what is scattering vs. extinction
        self.guess_scattering()

    def transmit(self, spectrum):
        '''
        Calculate the spectrum of light that results from
        transmitting a known light source through this
        atmosphere.

        Parameters
        ----------
        spectrum : Spectrum
            Any light source that you want to transmit.
        '''

        # create a composite object, linking the source and the atmosphere
        return Sunset(source=spectrum, atmosphere=self)

class DiscreteAtmosphere(Atmosphere):
    '''
    The DiscreteAtmosphere represents an atmosphere
    whose transmission as a function of wavelength
    can be represented as a grid of zenithward optical
    depths. Most pre-calculated models will fall
    into this category.
    '''

    def __init__(self, zenith_angle=0.0*u.deg, altitude=0.0, **kwargs):
        '''
        In inherited classes, this initialization relies on the
        existence of a method called ".read_transmission"
        that will create the attributes:
            ._wavelengths (with units of wavelength)
            ._tau_zenith_reference (unitless)
            .H (with units of length)
            .radius (with units of length)
        '''

        # read the transmission spectrum data
        self.read_transmission(**kwargs)
        self.default_wavelengths = self._wavelength

        # set the zenith angle (or fall back to the current setting)
        self.set_zenith_angle(zenith_angle)

        # set the altitude at which we're hovering
        self.set_altitude(altitude)


    def guess_scattering(self, visualize=False):
        '''
        Make a guess at the Rayleigh scattering component of
        the atmospheric extinction, for use in estimating
        the sky color.

        This should be called once per every new altitude.
        '''

        # figure out the appropriate normalization
        self._rayleigh_normalization = np.min(self.tau_zenith*self.default_wavelengths**4)

        w = self.default_wavelengths
        self.tau_zenith_scatter = self._rayleigh_normalization/w**4
        self.tau_zenith_absorb =  self.tau_zenith - self.tau_zenith_scatter

        if visualize:

            plt.plot(w, self.tau_zenith, zorder=100, label='total')
            plt.plot(w, self.tau_zenith_scatter, alpha=0.5, label='scattering')
            plt.plot(w, self.tau_zenith_absorb, alpha=0.5, label='absorbtion')
            plt.legend()
            plt.xscale('log')
            plt.yscale('log')
            plt.xlabel('Wavelength ({})'.format(w.unit.to_string('latex_inline')))
            plt.ylabel(r'$\tau_{zenith}$')

    """
    def tau_zenith_scatter(self):
        '''
        Estimate the optical depth to *absorption*.
        (at the default wavelengths of the grid)
        '''
        w = self.default_wavelengths
        return self._rayleigh_normalization/w**4

    def tau_zenith_absorb(self):
        '''
        Estimate the optical depth to *absorption*.
        (at the default wavelengths of the grid)
        '''
        # calculate the extinction from absorption
        return self.tau_zenith - self.tau_zenith_scatter()
    """

    def fortney_factor(self):
        '''
        The ratio of optical depth between a horizontal slant path
        through an atmosphere to a vertical path, accounting
        for the sphericity of the planet. This is a simple approach
        to connect exoplanet transmission spectra to vertical
        optical depths through an atmosphere (see Fortney 2005).

        Returns
        -------
        fortney_factor : float
            The ratio of slant to vertical optical depths.
        '''

        # calculate the fortney factor = ratio of slant/vertical optical depths
        return np.sqrt(2*np.pi*self.radius/self.H).decompose()

    def transit_radius(self, wavelength=None):
        '''
        Calculate the effective transit radius of the planet's atmosphere,
        as seen from infinitely far away.

        Parameters
        ----------
        wavelength : astropy.units.quantity.Quantity
            The wavelengths on which we want the transit radius.

        Returns
        -------
        transit_radius : astropy.units.quantity.Quantity
            The projected radius of the planet, with units.
        '''
        # make sure at least some grid of wavelengths is defined
        w = self.wavelength(wavelength)

        # figure out the slant optical depth at the reference radius
        effective_airmass = self.fortney_factor()
        tau = self._tau_zenith_reference*effective_airmass

        # bin this spectrum to the particular wavelength grid
        # KLUDGE -- this should really be in transmission space!
        neww, tau_slant = bintogrid(self._wavelength.to('nm').value, tau,
                               newx=w.to('nm').value,
                               drop_nans=False)

        # make sure the wavelengths match up
        assert(np.all(neww == w.to('nm').value))

        # convert back to tau
        z_over_H = np.log(tau_slant)

        return self.radius + self.H*z_over_H

    def transmission(self, wavelength=None, zenith_angle=None, altitude=None):
        '''
        Calculate the transmission through the atmosphere.

        Parameters
        ----------
        wavelength : astropy.units.quantity.Quantity
            The wavelengths on which we want the transmission.

        zenith_angle : astropy.units.quantity.Quantity
            The angle from zenith along which transmission
            should be calculated.

        altitude : float
            The altitude at which we should be floating
            in the atmosphere, in units of scale heights.

        Returns
        -------
        transmission : numpy.ndarray
            The fractional transmission through the atmosphere.
        '''

        # update the zenith angle, if necessary
        if zenith_angle is not None:
            self.set_zenith_angle(zenith_angle)

        # update the altitude, if necessary
        if altitude is not None:
            self.set_altitude(altitude)

        # make sure at least some grid of wavelengths is defined
        w = self.wavelength(wavelength)

        # figure out the transmission at this altitude
        # FIXME -- this is a major kludge! do the integral!
        effective_airmass = np.minimum(self.fortney_factor()/2.0, self.airmass)
        tau = self.tau_zenith*effective_airmass

        # bin this spectrum to the particular wavelength grid
        # FIXME: binning choice gets real scary with transmission
        neww, newt = bintogrid(self._wavelength.to('nm').value, np.exp(-tau),
                         newx=w.to('nm').value,
                         drop_nans=False)

        # make sure the wavelengths match up
        assert(np.all(neww == w.to('nm').value))

        # return the binned transmission
        return newt



from .imports import *
from .colortools import plot_rainbow
def setup_axes_with_rainbow(ax=None, rainbow=True, figsize=(5, 2.5)):
    # create new ax(s), unless we're supposed to over plot on one
    if ax is None:
        plt.figure(figsize=figsize)
        # decide whether to add a horizontal rainbow cartoon
        if rainbow:
            # create a two-part grid
            gs = GridSpec(  2, 1,
                            height_ratios=[0.05, 1],
                            hspace=0.0)


            # plot a cartoon rainbow, in a box above
            ax_rainbow = plt.subplot(gs[0])
            plt.sca(ax_rainbow)
            plot_rainbow(axes=ax_rainbow)

            ax_rainbow.get_yaxis().set_visible(False)
            ax_rainbow.get_xaxis().set_visible(False)
            ax_rainbow.set_facecolor('black')
            # (kludge to ensure black background for rainbow)
            if plt.gcf().get_facecolor() == (0.0, 0.0, 0.0, 1.0):
                plt.axis('off')

            # create the main ax
            ax = plt.subplot(gs[1], sharex=ax_rainbow)

        else:
            # simply create one simple ax
            ax = plt.subplot()


    # make sure we point back at the first plot
    plt.sca(ax)
    return ax



import numpy as np
import matplotlib.pyplot as plt
import copy, pkg_resources, os

from astropy.io import ascii
import astropy.units as u
import astropy.constants as con
from astropy.visualization import quantity_support

from .resampletools import bintogrid

from tqdm import tqdm

from matplotlib.gridspec import GridSpec, GridSpecFromSubplotSpec
from scipy.integrate import quad

data_directory = pkg_resources.resource_filename('rainbowconnection', 'data')



from .thermal import *

class Sun(Thermal):

    def __init__(self):
        self.radius = 1*u.Rsun
        self.teff = ((u.Lsun/(con.sigma_sb*4*np.pi*u.Rsun**2))**(1.0/4.0)).to(u.K)

        filename = os.path.join(data_directory, 'solarspectrum.txt')
        d = ascii.read(filename, comment='#')
        self._wavelength = d['wavelength'].data
        self._flux = d['irradiance'].data
        self.default_wavelengths = self._wavelength*u.nm



    def surface_flux(self, wavelength=None):

        # make sure at least some grid of wavelengths is defined
        w = self.wavelength(wavelength)

        # bin this spectrum to the particular wavelength grid
        neww, newf = bintogrid(self._wavelength,
                         self._flux,
                         newx=w.to('nm').value,
                         drop_nans=False)

        # make sure the wavelengths match up
        assert(np.all(neww == w.to('nm').value))

        # convert back to the surface flux from the Sun
        factor = ((1*u.au)**2/(1*u.Rsun)**2).decompose()
        return factor*newf*u.W/u.nm/u.m**2

    def integrate(self, lower=None, upper=None):
        '''
        Integrate the spectrum over wavelength.

        It gives a number with units identical to the results of
        `.spectrum()` but without the wavelength (W or W/m**2).

        Parameters
        ----------
        lower : astropy.units.quantity.Quantity
            The lower wavelength limit.

        upper : astropy.units.quantity.Quantity
            The lower wavelength limit.

        Returns
        -------
        integral : astropy.units.quantity.Quantity
            The integral over wavelength.
        '''

        # if wavelength limits are used, revert back to the numerical integral
        if (lower is not None) or (upper is not None):
            return super().integrate(lower=lower, upper=upper)

        # if there are infinite wavelength limits, use bolometric
        return 1*u.Lsun/self.normalization()



from .spectrum import *

class Thermal(Spectrum):
    def __init__(self, teff=5800*u.K, radius=1*u.Rsun):

        self.teff = teff
        self.radius = radius

    def intensity(self, wavelength):
        '''
        This function calculates the thermal emission intensity spectrum of a surface.

            Inputs:
                wavelength = numpy array of wavelengths (with astropy units)

            Outputs:
                Returns an array of thermal emission intensities,
                in astropy units of W/(m^2*micron*sr). This is a flux, which has
                already been integrated over solid angle.
        '''

        temperature = self.teff

        # define variables as shortcut to the constants we need
        h = con.h
        k = con.k_B
        c = con.c

        # this is the thing that goes into the exponent (it's units better cancel!)
        up = h*c/(wavelength*k*temperature)

        # calculate the intensity from the Planck function
        intensity = (2*h*c**2/wavelength**5/(np.exp(up) - 1))/u.steradian

        # return the intensity
        return intensity.to('W/(m**2*nm*sr)')

    def surface_flux(self, wavelength):
        '''
        This function calculates the thermal emission flux spectrum of a surface.

            Inputs:
                wavelength = numpy array of wavelengths (with astropy units)
                temperature = a single number, the temperature (with astropy units)

            Outputs:
                Returns an array of thermal emission fluxes,
                in astropy units of W/(m^2*micron). This is a flux, which has
                already been integrated over solid angle.
        '''

        # calculate the flux, knowing the angle integral will be pi steradians (for isotropic emission)
        flux = self.intensity(wavelength)*np.pi*u.steradian

        # return the flux, in convenient units
        return flux.to('W/(nm * m**2)')

    def __repr__(self):
        '''
        How should this object appear as a string?

        Returns
        -------
        s : str
            A simple string representation.
        '''

        basic = f'{self.__class__.__name__} ({self.teff:.0f}, {self.radius})'
        try:
            assert(self.distance is not None)
            return basic + f' at {self.distance}'
        except (AssertionError, AttributeError):
            return basic

    def integrate(self, lower=None, upper=None):
        '''
        Integrate the spectrum over wavelength.

        It gives a number with units identical to the results of
        `.spectrum()` but without the wavelength (W or W/m**2).

        Parameters
        ----------
        lower : astropy.units.quantity.Quantity
            The lower wavelength limit.

        upper : astropy.units.quantity.Quantity
            The lower wavelength limit.

        Returns
        -------
        integral : astropy.units.quantity.Quantity
            The integral over wavelength.
        '''

        # if wavelength limits are used, revert back to the numerical integral
        if (lower is not None) or (upper is not None):
            return super().integrate(lower=lower, upper=upper)

        # if there are infinite wavelength limits, do the integral analytically
        surface_flux = con.sigma_sb*self.teff**4
        factor = (self.surface_area()/self.normalization()).decompose()
        return factor*surface_flux



from ..imports import *
from ..colortools import plot_rainbow, rainbow_spectrum, CMFs, SpectralDistribution
import colour
from ..units import determine_quantity
from ..plottingtools import setup_axes_with_rainbow


bgkw = dict(color='gray', alpha=0.5, zorder=-100)

class Spectrum:
    '''
    The Spectrum class is a generic representation of the light from
    some emitting object, particularly for spherically symmetric emission.

    By default, the `.spectrum(wavelength)` method will return the
    spectral luminosity of the object. This is a quantity with units
    like W/nm, and it can be integrated over wavelength to provide
    the total luminosity of the light-emitting object, in W.

    The `.at(distance)` method creates an object representing the spectral
    flux from the object if seen from some particular distance. This is
    a quantity with units like W/nm/m**2, and it can be integrated over
    wavelength to provide the total flux of the light, in W/m**2.

    Classes that inherit from this will likely modify (at least) the
    surface_flux and surface_area methods.
    '''

    # the default grid of wavelengths
    default_wavelengths = np.arange(200, 1000)*u.nm

    def surface_flux(self, wavelength=None):
        '''
        The surface flux of the light source,
        in units like W/nm/m**2.

        (This should likely by overwritten by inheriting classes.)

        Parameters
        ----------
        wavelength : astropy.units.quantity.Quantity
            The wavelengths on which we want the flux.

        Returns
        -------
        surface_flux : astropy.units.quantity.Quantity
            The flux from the surface, usually in W/nm/m**2.
        '''

        # make sure a wavelength grid is defined
        w = self.wavelength(wavelength)

        # make a boring spectrum
        shape = np.shape(w)
        return np.ones(shape)*u.W/u.nm/u.m**2

    def surface_area(self):
        '''
        The surface area of the light source,
        in units like m**2.

        Returns
        -------
        surface_area : astropy.units.quantity.Quantity
            The emitting area of the surface, usually in m**2.
        '''
        return 4*np.pi*self.radius**2


    def wavelength(self, wavelength=None):
        '''
        A wrapper to ensure at least some grid of wavelengths
        gets defined. A default grid will be assumed, unless
        any wavelength array at all is passed.
        '''

        # make sure at least some wavelengths are defined
        if wavelength is None:
            wavelength = self.default_wavelengths
        return wavelength.to('nm')

    def spectrum(self, wavelength=None):
        '''
        The spectrum of the light source, as spectral luminosity (W/nm)
        or if a distance is defined as spectral flux (W/nm/m**2).

        Parameters
        ----------
        wavelength : astropy.units.quantity.Quantity
            The wavelengths on which we want the spectrum.

        Returns
        -------
        spectrum : astropy.units.quantity.Quantity
            The luminosity (W/nm) or flux (W/nm/m**2).
        '''

        # simplify the factor as best we can
        factor = (self.surface_area()/self.normalization()).decompose()

        # return the surface flux with appropriate normalization
        return factor*self.surface_flux(wavelength)

    def normalization(self):
        '''
        The normalization by which this Spectrum
        will be divided. It's flexible, to allow
        either spectral luminosity or spectral flux.

        Returns
        -------
        norm : astropy.units.quantity.Quantity
            A normalization
        '''
        try:
            # if there's a distance, return a flux
            assert(self.distance is not None)
            return 4*np.pi*self.distance**2
        except (AttributeError, AssertionError):
            # by default, simply return a luminosity
            return 1.0

    def angular_size(self):
        '''
        The angular size of the light source (if viewed from a distance).
        '''

        try:
            # if there's a distance, return an angular size
            assert(self.distance is not None)
            return np.arctan(self.radius/self.distance).to('deg')
        except (AttributeError, AssertionError):
            # complain if no distance is defined
            raise ValueError('''
            This Spectrum has no .distance attribute.
            Please consider using `.at(distance)` to
            create a new light source as viewed from
            a distance.
            ''')

    def at(self, distance=1*u.au):
        '''
        Create a new Spectrum representing the current
        light source viewed from some distance
        (assuming spherical symmetry).

        Parameters
        ----------
        distance : astropy.units.quantity.Quantity
            The distance at which we're viewing this source.

        Returns
        -------
        flux : Spectrum
            A new Spectrum, with the distance attached.
        '''

        # create a copy of the current spectrum
        new = copy.deepcopy(self)

        # update this copy's distance and return
        new.distance = distance
        return new

    # FIXME -- for analytic functions, it'd help to define some kind of
    # a bounding box in wavelength space, so this integral could be done
    # analytically or with scipy.integrate.quad
    def integrate(self, lower=None, upper=None):
        '''
        Integrate the spectrum over wavelength.

        It gives a number with units identical to the results of
        `.spectrum()` but without the wavelength (W or W/m**2).

        Parameters
        ----------
        lower : astropy.units.quantity.Quantity
            The lower wavelength limit.

        upper : astropy.units.quantity.Quantity
            The lower wavelength limit.

        Returns
        -------
        integral : astropy.units.quantity.Quantity
            The integral over wavelength
        '''

        w = self.default_wavelengths
        f = self.spectrum(w)


        ok = np.ones(np.shape(w)).astype(np.bool)
        if lower is not None:
            ok *= w >= lower

        if upper is not None:
            ok *= w <= upper
            #raise NotImplementedError('Wavelength limits not yet OK.')


        return np.trapz(f[ok], w[ok])

        #np.trapz(f.value, w.value)*f.unit*w.unit
        #wlower = lower or self.default_wavelengths[0]
        #wupper = upper or self.default_wavelengths[-1]
        #return quad(self.spectrum, wlower, wupper)



    def to_sd(self):
        '''
        Create a `colour` SpectralDistribution from this object,
        solely covering the visible range. This can be used
        to estimate the true color of this spectrum.
        '''

        # pick wavelengths directly from the color-matching functions
        w = CMFs.wavelengths*u.nm

        # normalize only within the visible range
        ok = (w < 700*u.nm) & (w > 390*u.nm)
        w = w[ok]
        # (note: this is a kludge to avoid making spectra
        #  with most of their luminosity outside the visible
        #  appear as really dark and dull. there is probably
        #  a much cleverer way of making sure the normalization
        #  does something reasonable.)

        # calculate the spectrum at those wavelengths
        f = self.spectrum(w)

        # create the spectral distribution
        sd = SpectralDistribution(dict(zip(w.value, f.value/np.max(f.value))))

        # return it
        return sd

    def to_color(self):
        '''
        Determine the RGB color of this spectrum.

        Returns
        -------
        rgb : numpy.ndarray
            3-element array containing RGB values
            that can be fed into matplotlib.
        '''

        # create a colour SpectralDistribution
        sd = self.to_sd()

        # convert to XYZ
        # (maybe use `k=` option for relative scaling between sources?)
        XYZ = colour.sd_to_XYZ(sd)
        if (np.min(XYZ) < 0) or np.max(XYZ) > 100:
            print(f'XYZ={XYZ} is outside of [0, 100]!')

        # convert to RGB
        RGB = colour.XYZ_to_sRGB(XYZ / 100)
        if (np.min(RGB) < 0) or np.max(RGB) > 1:
            pass
            #print(f'RGB={RGB} is outside of [0, 1]!')

        # trim out underenderable colors (this is sneaky!)
        #clipped_RGB = np.maximum(0, np.minimum(1, RGB))
        clipped_RGB = np.maximum(0, RGB)
        # instead of clip, should we add to everything until we get to zero?

        # kludge to maximize brightness, for every color!
        clipped_RGB /= np.max(clipped_RGB)
        return clipped_RGB

    def __repr__(self):
        '''
        How should this object appear as a string?

        Returns
        -------
        s : str
            A simple string representation.
        '''
        try:
            assert(self.distance is not None)
            return f'{self.__class__.__name__} at {self.distance}'
        except (AssertionError, AttributeError):
            return f'{self.__class__.__name__}'

    def set_power(self, power=100*u.W):
        '''
        Change the radius of the object so that it will
        emit a specified power, with the same spectral shape.

        Parameters
        ----------
        power : astropy.units.quantity.Quantity
            The total power we want the object to emit.
        '''

        # calculate the total luminosity of this object
        total = self.integrate()

        # make sure we're dealing with an actual luminosity
        assert(total.unit.is_equivalent('W'))

        # calculation a new normalization
        normalization = (power/total).decompose()

        # change the radius of this object
        self.radius *= np.sqrt(normalization)
        self.power = power

    def mean_intensity(self, wavelength=None):
        '''
        Calculate the mean intensity field created by the source.

        The mean intensity field represents the intensity
        of the source, smeared over a full 4pi steradians.

        (This makes sense only for spectra viewed from a distance.)
        '''

        # what is the intensity of the disk
        F_disk = self.spectrum(wavelength)


        # make sure we're dealing with a flux
        assert(F_disk.unit.is_equivalent(u.W/u.m**2/u.nm))

        # calculate the mean intensity
        solid_angle = np.pi*self.angular_size()**2

        # calculate the mean intensity field
        J = F_disk/4/np.pi/u.sr

        return J

    def disk_intensity(self, wavelength=None):
        '''
        Calculate the intensity of the disk of the source.

        The disk intensity represents the intensity of staring
        directly at the disk. The flux of the source is its
        intensity integrated over solid angle, so two sources
        with the same intensities can have very different
        fluxes, based on their apparent angular sizes.
        For example, the Sun seen `.at` different distances
        will have different fluxes but the same disk intensity.

        (This makes sense only for spectra viewed from a distance.)
        '''

        # what is the intensity of the disk
        F_disk = self.spectrum(wavelength)


        # make sure we're dealing with a flux
        assert(F_disk.unit.is_equivalent(u.W/u.m**2/u.nm))

        # calculate the mean intensity
        solid_angle = np.pi*self.angular_size()**2
        I_disk = F_disk/solid_angle

        return I_disk

    # ===================================
    #
    #        PLOTTING METHODS
    #
    # ===================================

    def plot(self,  ax=None,
                    wavelength=None,
                    rainbow=True,
                    color='auto',
                    style='dark_background',
                    figsize=(5, 2.5),
                    **kwargs):
        '''
        A quick tool to plot a spectrum.

        Parameters
        ----------
        ax : matplotlib.axes._subplots.AxesSubplot
            Specify an axes object into which
            this plot should be drawn. This allows
            overplotting multiple spectra, for example
            with a structure like
                ```
                ax = Thermal(teff=5800*u.K, radius=1*u.Rsun).plot()
                ax = Sun.plot(ax)
                ```
            To overplot on current axes use `ax=plt.gca()`.

        wavelength : astropy.units.quantity.Quantity
            A grid of wavelengths on which the spectrum should
            be plotted. If None, the function defaults to
            covering visible wavelengths at 1nm resolution.

        rainbow : bool
            Should we add an extra rainbow above the plot,
            to indicate how wavelengths match to visible light?

        color : str
            The color for drawing the spectrum.
            'auto' represents the actual visible color.

        '''

        # set up to use a dark background for the plot; make sure units match
        with plt.style.context(style), quantity_support():

            # setup the basic axes
            ax = setup_axes_with_rainbow(ax=ax, rainbow=rainbow, figsize=figsize)

            # make sure at least some wavelengths are defined
            w = self.wavelength(wavelength)

            # pull out the spectrum
            f = self.spectrum(w)

            # plot the spectrum
            if color == 'auto':
                color = self.to_color()
                background = ax.get_facecolor()[0:3]
                if np.max(color - background) < 0.05:
                    print(f'''
                    The inferred color {color} might be a little
                    too close to {background} to be visible. Consider
                    plotting without the `color='auto'` option.
                    ''')
            plt.plot(w, f, color=color, label=self, **kwargs)

            # add the axis labels
            wunit = w.unit.to_string('latex_inline')
            funit = f.unit.to_string('latex_inline')
            plt.xlabel(f'Wavelength ({wunit})')
            plt.ylabel(f'{determine_quantity(f.unit)} ({funit})')

        return ax


    def plot_rgb(self,  ax=None,
                        wavelength=None,
                        rainbow=True,
                        foreground=True,
                        style='dark_background',
                        figsize=(5, 2.5),
                        **kwargs):
        '''
        A quick tool to plot a spectrum, just as RGB bars.

        Parameters
        ----------
        ax : matplotlib.axes._subplots.AxesSubplot
            Specify an axes object into which
            this plot should be drawn. This allows
            overplotting multiple spectra, for example
            with a structure like
                ```
                ax = Thermal(teff=5800*u.K, radius=1*u.Rsun).plot()
                ax = Sun.plot(ax)
                ```
            To overplot on current axes use `ax=plt.gca()`.

        wavelength : astropy.units.quantity.Quantity
            A grid of wavelengths on which the spectrum should
            be plotted. If None, the function defaults to
            covering visible wavelengths at 1nm resolution.

        rainbow : bool
            Should we add an extra rainbow above the plot,
            to indicate how wavelengths match to visible light?

        color : str
            The color for drawing the spectrum.
            'auto' represents the actual visible color.

        '''

        # set up to use a dark background for the plot; make sure units match
        with plt.style.context(style), quantity_support():

            # setup the basic axes
            ax = setup_axes_with_rainbow(ax=ax, rainbow=rainbow, figsize=figsize)

            # make sure at least some wavelengths are defined
            w = self.wavelength(wavelength)

            rgb = self.to_color()

            blue = [400, 495]
            green = [495, 590]
            red = [590, 685]

            centers = [np.mean(c) for c in [red, green, blue]]
            widths = [c[1] - c[0] for c in [red, green, blue]]

            if foreground:
                kw = dict(color = [np.array([1, 0, 0]),
                                   np.array([0, 1, 0]),
                                   np.array([0, 0, 1])],
                          edgecolor='white',
                          zorder = 0)
            else:

                kw = dict(**bgkw)
                kw['edgecolor'] = 'none'
            plt.bar(centers, rgb*100, widths, **kw) # linewidth=2,


            # add the axis labels
            wunit = w.unit.to_string('latex_inline')
            plt.xlabel(f'Wavelength ({wunit})')
            plt.ylabel(f'Brightness (%)')

        return ax

    def plot_as_rainbow(self,  ax=None,
                                rainbow=True,
                                color='auto',
                                style='dark_background',
                                foreground=True,
                                ylim=[0, 120],
                                figsize=(5,2.5),
                                **kwargs):
        '''
        A quick tool to plot a spectrum, just as RGB bars.

        Parameters
        ----------
        ax : matplotlib.axes._subplots.AxesSubplot
            Specify an axes object into which
            this plot should be drawn. This allows
            overplotting multiple spectra, for example
            with a structure like
                ```
                ax = Thermal(teff=5800*u.K, radius=1*u.Rsun).plot()
                ax = Sun.plot(ax)
                ```
            To overplot on current axes use `ax=plt.gca()`.

        wavelength : astropy.units.quantity.Quantity
            A grid of wavelengths on which the spectrum should
            be plotted. If None, the function defaults to
            covering visible wavelengths at 1nm resolution.

        rainbow : bool
            Should we add an extra rainbow above the plot,
            to indicate how wavelengths match to visible light?

        color : str
            The color for drawing the spectrum.
            'auto' represents the actual visible color.

        '''

        # set up to use a dark background for the plot; make sure units match
        with plt.style.context(style), quantity_support():

            # setup the basic axes
            ax = setup_axes_with_rainbow(ax=ax, rainbow=rainbow, figsize=figsize)

            # make sure at least some wavelengths are defined
            w = self.wavelength(np.arange(330, 760, 1)*u.nm)
            f = self.spectrum(w)
            # KLUDGE?
            norm = np.max(f.value[(w > 400*u.nm) & (w < 685*u.nm)]) / 100
            if foreground:
                rainbow_spectrum(axes=ax, wavelength=w.to('nm').value, flux=f.value/norm,
                                 rainbowtop=np.max(ylim))
                plt.plot(w, f/norm, color='white') #, linewidth=2
            else:
                plt.fill_between(w, f/norm, linewidth=0, **bgkw)
            plt.ylim(*ylim)

            # add the axis labels
            wunit = w.unit.to_string('latex_inline')
            plt.xlabel(f'Wavelength ({wunit})')
            plt.ylabel(f'Brightness (%)')

        return ax

    def cartoon_sun(self):
        '''
        Create a simple cartoon of the star,
        with appropriate color and angular size.
        '''

        # figure out the color of the star
        rgb = self.to_color()

        with plt.style.context('dark_background'):
            fi = plt.figure(figsize=(5, 2.5))
            ax = fi.add_axes([0, 0, 1, 1])
            plt.scatter(0, 0, c=rgb, s=8000)
            plt.xticks([])
            plt.yticks([])
            plt.axis('off')



from ..imports import *
from ..sources import Spectrum
from ..animatetools import *
from .sky import *

class Sunset(Spectrum):

    def __init__(self, source, atmosphere):
        '''
        Initialize this composite object, connecting a light source
        and an atmosphere together into a Sunset calculator.

        Parameters
        ----------
        source : rainbowconnection.Spectrum
            The light source from above the atmosphere.
        atmosphere : rainbowconnection.Atmosphere
            The atmosphere through which the light source propagates.
        '''

        self.source = source
        self.atmosphere = atmosphere
        self.sky = Sky(self)

    # KLUDGE - there's got to be a better way to organize this?!
    @property
    def radius(self):
        return self.source.radius

    # KLUDGE - there's got to be a better way to organize this?!
    @property
    def distance(self):
        try:
            return self.source.distance
        except AttributeError:
            return None

    def __repr__(self):
        '''
        How should this be represented as a string?
        '''

        s = repr(self.source)
        t = repr(self.atmosphere)
        return f'{s}\n through\n{t}'

    def set_zenith_angle(self, *args, **kwargs):
        '''
        Pass zenith commands up to the atmosphere.
        '''
        self.atmosphere.set_zenith_angle(*args, **kwargs)

    def set_altiude(self, *args, **kwargs):
        '''
        Pass altitude commands up to the atmosphere.
        '''
        self.atmosphere.set_altitude(*args, **kwargs)

    def spectrum(self, wavelength=None, zenith_angle=None, altitude=None):
        '''
        The spectrum of the light source viewed through the atmosphere.

        Parameters
        ----------
        wavelength : astropy.units.quantity.Quantity
            The wavelengths on which we want the spectrum.

        zenith_angle : astropy.units.quantity.Quantity
            The angle away from zenith along which
            the transmission of the atmosphere should
            be calculated.

        altitude : float
            The altitude at which we should be floating
            in the atmosphere, in units of scale heights.

        Returns
        -------
        spectrum : astropy.units.quantity.Quantity
            The luminosity (W/nm) or flux (W/nm/m**2).
        '''

        # the transmission wavelength sets the grid
        w = self.atmosphere.wavelength(wavelength)

        # get the transmission a
        t = self.atmosphere.transmission(wavelength=w,
                                         zenith_angle=zenith_angle,
                                         altitude=altitude)

        # get the original spectrum
        s = self.source.spectrum(wavelength=w)

        # return the transmitted spectrum
        return s*t

    def transit_depth(self, wavelength=None):
        '''
        The transit depth of the planet, passing in front of the light source.

        Parameters
        ----------
        wavelength : astropy.units.quantity.Quantity
            The wavelengths on which we want the spectrum.

        Returns
        -------
        transitdepth : np.ndarry
            The fraction of starlight blocked, unitless.
        '''

        # the transmission wavelength sets the grid
        w = self.atmosphere.wavelength(wavelength)

        # get the transmission a
        rp = self.atmosphere.transit_radius(wavelength=w)

        # get the original spectrum
        rs = self.source.radius

        return (rp/rs).decompose()**2


    def plot_disk(self, ax=None, zenith_angle=89*u.deg, azimuth_angle=0*u.deg):
        '''
        Plot the disk of the star.

        Parameters
        ----------
        ax : matplotlib.axes._subplots.AxesSubplot
            The axes into which this plot should be drawn.
        zenith_angle : astropy.units.quantity.Quantity
            The angle away from zenith.
        azimuth_angle : astropy.units.quantity.Quantity
            The azimuth angle along the horizon.
        '''

        with plt.style.context('dark_background'), quantity_support():

            if ax is not None:
                plt.sca(ax)

            # set the zenith angle
            self.set_zenith_angle(zenith_angle)

            # figure out the color of the star
            rgb = self.to_color()

            c = plt.Circle(xy=[azimuth_angle, 90*u.deg - zenith_angle],
                           radius=self.source.angular_size(),
                           facecolor=rgb, zorder=1000)
            plt.gca().add_patch(c)

            size = self.source.angular_size()*1.1
            plt.xlim(azimuth_angle - size, azimuth_angle + size)
            plt.ylim(90*u.deg - zenith_angle - size,
                     90*u.deg - zenith_angle + size)
            plt.axis('scaled')
            return c

    def plot_sky(self, maxelevation=20*u.deg,
                       minelevation=0*u.deg,
                       skyresolution=0.5*u.deg,
                       skynormalization=0.7):
        '''
        Plot a diffuse sky behind the disk of the star.

        Parameters
        ----------
        maxelevation : astropy.units.quantity.Quantity
            The maximum elevation above the horizon.
        minelevation : astropy.units.quantity.Quantity
            The minimum elevation above the horizon.
        skyresolution : astropy.units.quantity.Quantity
            The vertical extent of constant-color stripes
            when rendering the atmosphere. Stripes larger
            than ~1/20 the vertical extent will be very
            noticeable.
        skynormalization : float
            By what factor do we down-weight the intensity
            of the sky. To be quantitatively correct, this
            should be something like the ratio of the
            stellar disk intensity to the diffuse sky
            brightness, but that factor is so tiny that
            the sky would always seem to be black.
        '''

        if self.atmosphere.__class__.__name__== 'NoAtmosphere':
            return

        # set the elevations where we will calculate sky colors
        elevations = np.arange(minelevation.to('deg').value,
                               (maxelevation + skyresolution).to('deg').value,
                               skyresolution.to('deg').value)*u.deg

        # calculate a list of sky colors, one for each elevation
        colors = []
        for e in elevations:

            # point at a particular stripe of the sky
            self.sky.set_zenith_angle(90*u.deg - e)

            # estimate the color of that stripe
            colors.append(self.sky.to_color()*skynormalization)

        # draw horizontal bars with these colors
        plt.barh(y=elevations,
                 width=20,
                 height=skyresolution,
                 left=-10,
                 zorder=-100,
                 color=colors)

    def plot_sunset(self, ax=None,
                          zenith_angle=89*u.deg,
                          azimuth_angle=0*u.deg,
                          maxelevation=20*u.deg,
                          minelevation=-5*u.deg,
                          skyresolution=0.5*u.deg,
                          skynormalization=0.7):

        '''
        Plot the disk of the star and the diffuse sky.

        Parameters
        ----------
        ax : matplotlib.axes._subplots.AxesSubplot
            The axes into which this plot should be drawn.
        zenith_angle : astropy.units.quantity.Quantity
            The angle away from zenith.
        azimuth_angle : astropy.units.quantity.Quantity
            The azimuth angle along the horizon.
        maxelevation : astropy.units.quantity.Quantity
            The maximum elevation above the horizon.
        minelevation : astropy.units.quantity.Quantity
            The minimum elevation above the horizon.
        skyresolution : astropy.units.quantity.Quantity
            The vertical extent of constant-color stripes
            when rendering the atmosphere. Stripes larger
            than ~1/20 the vertical extent will be very
            noticeable.
        skynormalization : float
            By what factor do we down-weight the intensity
            of the sky. To be quantitatively correct, this
            should be something like the ratio of the
            stellar disk intensity to the diffuse sky
            brightness, but that factor is so tiny that
            the sky would always seem to be black.
        '''

        # make sure we're using a dark background and astropy units
        with plt.style.context('dark_background'), quantity_support():

            if ax is not None:
                plt.sca(ax)

            # put the sun at a particular elevation
            self.set_zenith_angle(zenith_angle)

            # plot the diffuse sky
            self.plot_sky(maxelevation=maxelevation,
                          minelevation=minelevation)

            # plot the disk of the sun
            self.plot_disk(zenith_angle=zenith_angle,
                           azimuth_angle=azimuth_angle)

            # fuss with the axis labels
            plt.axis('scaled')
            plt.xlim(-5*u.deg, 5*u.deg)
            plt.ylim(0*u.deg, maxelevation)
            plt.ylabel('Angle above Horizon')
            ax = plt.gca()
            ax.get_xaxis().set_visible(False)
        return ax

    def animate_sunset(self, filename='sunset.mp4',
                             maxelevation=30*u.deg,
                             skyresolution=0.5*u.deg,
                             skynormalization=0.7,
                             motionresolution=0.5*u.deg):
        '''
        Animate the sun setting.

        Parameters
        ----------
        filename : string
            The filename where this animation should be saved
        maxelevation : astropy.units.quantity.Quantity
            The maximum elevation above the horizon.
        skyresolution : astropy.units.quantity.Quantity
            The vertical extent of constant-color stripes
            when rendering the atmosphere. Stripes larger
            than ~1/20 the vertical extent will be very
            noticeable.
        skynormalization : float
            By what factor do we down-weight the intensity
            of the sky. To be quantitatively correct, this
            should be something like the ratio of the
            stellar disk intensity to the diffuse sky
            brightness, but that factor is so tiny that
            the sky would always seem to be black.
        motionresolution : astropy.units.quantity.Quantity
            The angular step the star moves between
            frames of the animation.
        '''


        # get the appropriate animation writer
        wri = get_writer(filename)

        # calculate the grid of zenith angles for the star
        min_zenith = 90*u.deg - maxelevation - self.source.angular_size()
        minelevation = -(self.source.angular_size() + motionresolution)
        max_zenith = 90*u.deg - minelevation
        zenith_angles = np.arange(min_zenith.to('deg').value,
                                  max_zenith.to('deg').value,
                                  motionresolution.to('deg').value)*u.deg


        # make a figure with a dark background
        with plt.style.context('dark_background'), quantity_support():

            # create a new figure
            fi, ax = plt.subplots(1,1)

            # save to frames of an animation
            with wri.saving(fi, filename, 400):

                # loop over zenith angles
                for z in tqdm(zenith_angles):

                    # clear whatever was in the previous axes
                    ax.cla()

                    # plot the sunset at this current stellar zenith angle
                    self.plot_sunset(ax=ax,
                                     zenith_angle=z,
                                     maxelevation=maxelevation,
                                     minelevation=minelevation)

                    # grab this frame in the animation
                    wri.grab_frame()

    def animate_everything(self, filename='everything-sunset.mp4',
                                 ingredients=['sky', 'sky-zoom', 'rgb', 'spectrum'],
                                 maxelevation=50*u.deg,
                                 skyresolution=0.5*u.deg,
                                 skynormalization=0.7,
                                 motionresolution=0.5*u.deg,
                                 **kwargs):

        '''
        Make an animation that summarizes
        lots of information about a sunset.

        Parameters
        ----------
        filename : string
            The filename where this animation should be saved
        ingredients : list
            The components to draw. These are:
                'sky' = a slice of many elevations
                'sky-zoom' = just the disk of the star
                'rgb' = a plot of relative RGB colors
                'spectrum' = a plot of the spectrum
            Any subset of these is allowed.
        maxelevation : astropy.units.quantity.Quantity
            The maximum elevation above the horizon.
        skyresolution : astropy.units.quantity.Quantity
            The vertical extent of constant-color stripes
            when rendering the atmosphere. Stripes larger
            than ~1/20 the vertical extent will be very
            noticeable.
        skynormalization : float
            By what factor do we down-weight the intensity
            of the sky. To be quantitatively correct, this
            should be something like the ratio of the
            stellar disk intensity to the diffuse sky
            brightness, but that factor is so tiny that
            the sky would always seem to be black.
        motionresolution : astropy.units.quantity.Quantity
            The angular step the star moves between
            frames of the animation.
        **kwargs : dict
            All extra keyword arguments will be passed to
            plot_everything.
        '''


        # get the appropriate animation writer
        wri = get_writer(filename)

        # calculate the grid of zenith angles for the star
        min_zenith = 90*u.deg - maxelevation - self.source.angular_size()
        minelevation = -(self.source.angular_size() + motionresolution)
        max_zenith = 90*u.deg - minelevation
        zenith_angles = np.arange(min_zenith.to('deg').value,
                                  max_zenith.to('deg').value,
                                  motionresolution.to('deg').value)*u.deg

        with plt.style.context('dark_background'), quantity_support():

            fi = self.plot_everything(zenith_angles[0], **kwargs)

            # save to frames of an animation
            with wri.saving(fi, filename, fi.get_dpi()):

                # loop over zenith angles
                for z in tqdm(zenith_angles):

                    # clear whatever was in the previous axes
                    fi.clf()

                    # plot the sunset at this current stellar zenith angle
                    self.plot_everything(fi=fi, zenith_angle=z,
                                         ingredients=ingredients, **kwargs)

                    # grab this frame in the animation
                    wri.grab_frame()


    def plot_everything(self, ingredients=['sky', 'sky-zoom', 'rgb', 'spectrum'],
                              zenith_angle=85*u.deg,
                              maxelevation=50*u.deg,
                              fi=None,
                              pixels=[1920, 1080],
                              width=8,
                              filename=None,
                              ):
        '''
        Make an animation that summarizes
        lots of information about a sunset.

        Parameters
        ----------
        ingredients : list
            The components to draw. These are:
                'sky' = a slice of many elevations
                'sky-zoom' = just the disk of the star
                'rgb' = a plot of relative RGB colors
                'spectrum' = a plot of the spectrum
            Any subset of these is allowed.
        zenith_angle : astropy.units.quantity.Quantity
            The angle away from zenith.
        maxelevation : astropy.units.quantity.Quantity
            The maximum elevation above the horizon
            (for setting the plot limits.)
        aspect : list
            The aspect ratio [width, height] of the
            size of the plot.
        '''

        with plt.style.context('dark_background'), quantity_support():
            xpixels, ypixels = pixels
            dpi = xpixels/width
            scale = 12/width
            if fi is None:

                fi = plt.figure(figsize=(width, ypixels*width/xpixels),
                                dpi=dpi)

            # set up the basic columns
            gs = GridSpec(  1, 4,
                            width_ratios=[0.18, 0.4, 0.09*scale,  0.9],
                            wspace=0.0)

            ax = {}

            # set up the contextualizing plots on the left
            gs_geometry = GridSpecFromSubplotSpec(2, 1,
                                                  height_ratios=[1, 1],
                                                  hspace=0.5*scale,
                                                  subplot_spec=gs[1])

            #ax['cartoon'] = plt.subplot(gs_geometry[0])
            if 'sky-zoom' in ingredients:
                ax['sky-zoom'] = plt.subplot(gs_geometry[0])
                self.plot_disk(ax=ax['sky-zoom'], zenith_angle=zenith_angle)
                width = 1.1*self.source.angular_size()
                plt.xlim(-width, width)
                plt.ylim(90*u.deg - zenith_angle - width, 90*u.deg - zenith_angle + width)
                ax['sky-zoom'].get_xaxis().set_visible(False)
                ax['sky-zoom'].get_yaxis().set_visible(False)
                plt.axis('scaled')
                plt.axis('off')

            # set up the main panel showing the sunset over the full sky
            if 'sky' in ingredients:
                ax['sky'] = plt.subplot(gs[0])
                # actually make the plots
                self.plot_sunset(ax=ax['sky'], zenith_angle=zenith_angle, maxelevation=maxelevation)

            #
            # set up the spectrum panels
            gs_spectra = GridSpecFromSubplotSpec(2, 1,
                                                 height_ratios=[1, 1],
                                                 subplot_spec=gs[3],
                                                 hspace=0.5*scale)
            if 'rgb' in ingredients:
                ax['rgb'] = plt.subplot(gs_spectra[0])
                self.plot_rgb(ax=ax['rgb'])
                plt.ylim(0, 120)
                plt.xlim(360*u.nm, 740*u.nm)

            if 'spectrum' in ingredients:
                ax['spectrum'] = plt.subplot(gs_spectra[1],
                                             sharex=ax['rgb'], sharey=ax['rgb'])
                self.plot_as_rainbow(ax['spectrum'])
                plt.ylim(0, 120)
                plt.xlim(360*u.nm, 740*u.nm)

        if filename is not None:
            plt.savefig(filename, facecolor='black')
        return fi

    def plot_rgb(self, ax=None, **kwargs):
        '''
        Wrapper to add the unextincted spectrum
        in dark gray in the background for context.
        '''
        ax = Spectrum.plot_rgb(self, ax=ax, **kwargs)
        self.source.plot_rgb(foreground=False, ax=ax)

    def plot_as_rainbow(self, ax=None, **kwargs):
        '''
        Wrapper to add the unextincted spectrum
        in dark gray in the background for context.
        '''
        ax = Spectrum.plot_as_rainbow(self, ax=ax, **kwargs)
        self.source.plot_as_rainbow(foreground=False, ax=ax)



from .atmosphere import *

class Exoplanet(DiscreteAtmosphere):

    def __init__(self, wavelength, planet_radius, H=20*u.km,
                       altitude=0.0,
                       zenith_angle=0.0*u.deg,
                       **kwargs):
        '''
        Initialize a generic exoplanet atmosphere from a model exoplanet
        transmission spectrum, specified by the inputs `wavelength`
        and `planet_radius`

        Parameters
        ----------
        wavelength : astropy.units.quantity.Quantity
            The wavelengths of the transmission spectrum,
            in units equivalent to [microns].
        planet_radius : astropy.units.quantity.Quantity
            The wavelength-dependent radius of the planet,
            in units equivalent to [m].
        H : astropy.units.quantity.Quantity
            The scale height of the atmosphere. This is used
            only for estimating how to transform between
            slant and zenith viewing geometries.
        altitude : float
            The altitude at which we should be floating
            in the atmosphere, in units of scale heights.
        zenith_angle : astropy.units.quantity.Quantity
            The default angle from zenith along which transmission
            should be calculated (this can be changed).

        '''


        # define the wavelength grid
        self._wavelength = wavelength.to('nm')

        # store the effective transit radius
        self._transit_radius = planet_radius

        # pick an (arbitrary-ish) reference radius
        reference_radius = np.min(self._transit_radius)

        # define geometry of the atmosphere (how spherical?)
        self.H = H
        self.radius = reference_radius

        # calculate the optical depth at zenith
        z_over_H = ((self._transit_radius - reference_radius)/self.H).decompose()
        tau_slant = np.exp(z_over_H)
        self._tau_zenith_reference = (tau_slant/self.fortney_factor()).decompose()


        # define the default wavelength grid to use
        self.default_wavelengths = self._wavelength

        # set the zenith angle (or fall back to the current setting)
        self.set_zenith_angle(zenith_angle)

        # set the altitude at which we're hovering
        self.set_altitude(altitude)



from .atmosphere import *

class NoAtmosphere(Atmosphere):
    '''
    A fake atmosphere that represents a completely transparent atmosphere.
    This might be useful as a placeholder, if you want to make
    plots/animations that directly compare the spectrum of a
    source with or without extinction by an atmosphere.
    '''

    def __repr__(self):
        '''
        How should this object appear as a string?

        Returns
        -------
        s : str
            A simple string representation.
        '''
        #H/R={scale},
        return f'{self.__class__.__name__}Atmosphere'


    def __init__(self, zenith_angle=0.0*u.deg, **kwargs):
        '''
        Initialize the fake atmosphereless atmosphere.
        '''

        # read the transmission spectrum data
        self.default_wavelengths = np.arange(300, 1000, 1)*u.nm

        # set the zenith angle (or fall back to the current setting)
        self.set_zenith_angle(zenith_angle)

    def transmission(self, wavelength=None, zenith_angle=None, **kw):
        '''
        Calculate the transmission through the atmosphere.

        Parameters
        ----------
        wavelength : astropy.units.quantity.Quantity
            The wavelengths on which we want the transmission.

        zenith_angle : astropy.units.quantity.Quantity
            The angle from zenith along which transmission
            should be calculated.

        Returns
        -------
        transmission : numpy.ndarray
            The fractional transmission through the atmosphere.
        '''

        # update the zenith angle, if necessary
        if zenith_angle is not None:
            self.set_zenith_angle(zenith_angle)

        # make sure at least some grid of wavelengths is defined
        w = self.wavelength(wavelength)

        # return 100% transmission at all wavelengths
        return np.ones(np.shape(w))



import matplotlib.animation as ani

def get_writer(filename, fps=30, **kw):
	'''
	Try to get an appropriate animation writer,
	given the filename provided.

	Parameters
	----------

	filename : str
		The output filename string for the animation.

	fps : float
		Frames/second.

	kw : dict
		All other keywords will be passed to the initialization
		of the animation writer.
	'''
	if '.mp4' in filename:
		try:
			writer = ani.writers['ffmpeg'](fps=fps, **kw)
		except (RuntimeError, KeyError):
			raise RuntimeError('This computer seems unable to ffmpeg.')
	else:
		try:
			writer = ani.writers['pillow'](fps=fps, **kw)
		except (RuntimeError, KeyError):
			writer = ani.writers['imagemagick'](fps=fps, **kw)
			raise RuntimeError('This computer seem unable to animate?')
	return writer
	



from .spectrum import *
from .thermal import Thermal

def gauss(x, x0, sigma):
    '''
    Simple Gaussian.
    '''
    return 1.0/np.sqrt(2*np.pi)/sigma*np.exp(-0.5*((x-x0)/sigma)**2)

class LightBulb(Spectrum):
    def __repr__(self):
        '''
        How should this object appear as a string?

        Returns
        -------
        s : str
            A simple string representation.
        '''

        basic = f'{self.__class__.__name__}LightBulb ({self.power})'
        try:
            assert(self.distance is not None)
            return basic + f' at {self.distance}'
        except (AssertionError, AttributeError):
            return basic

class Incandescent(LightBulb, Thermal):
    def __init__(self, power=100*u.W):

        Thermal.__init__(self, teff=3500*u.K, radius=1*u.mm)

        # renormalize that radius to match the light output
        self.set_power(power)


class Sodium(LightBulb):
    def __init__(self, power=100*u.W):

        # start with a ridiculous radius
        self.radius = 1*u.m

        # renormalize that radius to match the light output
        self.set_power(power)

    def surface_flux(self, wavelength):
        '''
        This function calculates a cartoon spectrum of a sodium lamp.

            Inputs:
                wavelength = numpy array of wavelengths (with astropy units)
                temperature = a single number, the temperature (with astropy units)

            Outputs:
                Returns an array of thermal emission fluxes,
                in astropy units of W/(m^2*micron). This is a flux, which has
                already been integrated over solid angle.
        '''

        # make sure a wavelength grid is defined
        w = self.wavelength(wavelength)
        flux = np.zeros(np.shape(w))/u.nm

        # set the two lines
        lines = np.array([588.9950, 589.5924])*u.nm
        width = 2.0*u.nm

        # give a normalization
        brightness = 1*u.W/u.m**2

        # add the two lines to the spectrum
        for l in lines:
            flux += gauss(w, x0=l, sigma=width)


        # return the flux, in convenient units
        return (brightness*flux).to('W/(nm * m**2)')



from ..imports import *
from ..sources import Spectrum, Thermal
from colour.phenomena.rayleigh import rayleigh_optical_depth

class Sky(Spectrum):

    def __init__(self, sunset):
        '''
        Initialize this composite object, connecting a light source
        and an atmosphere together into a diffuse Sky calculator.

        Parameters
        ----------
        sunset : rainbowconnection.Sunset
            Any object with both a source and an atmosphere.
            Calculations related to the diffuse sky brightness
            require both, so it's simplest just to pass
            a sunset object in which they are already linked.
        '''

        self.sunset = sunset
        self.source = sunset.source
        self.atmosphere = sunset.atmosphere

        try:
            self.B = Thermal(self.atmosphere.T).intensity
        except AttributeError:
            def f(wavelength):
                return np.zeros(np.shape(wavelength))*u.W/u.m**2/u.nm*u.sr
            self.B = f

        self.set_zenith_angle(0*u.deg)
        # self.is_earth = 'Earth' in self.atmosphere.__class__.__name__


    def set_zenith_angle(self, zenith_angle=0*u.deg):
        '''
        Set the angle from zenith (for a slice of the sky.)

        Parameters
        ----------
        zenith_angle : astropy.units.quantity.Quantity
            The angle away from zenith along which
            the transmission of the atmosphere should
            be calculated.
        '''

        # set the zenith angle and airmass
        self.zenith_angle = np.minimum(zenith_angle, 90*u.deg)
        self.airmass = 1/np.cos(self.zenith_angle)


    def spectrum(self, wavelength=None):
        '''
        Calculate intensity of the diffuse sky at a given zenith angle.
        Currently this accounts only for Rayleigh scattering (assuming
        perfect 1/wavelength**4) and for thermal emission (which is
        usually negligible below about 1700K). The relative contributions
        of scattering and aborption/emission are estimated in a very
        cartoonish fashion from the overall extinction, and assumes
        there is at least one wavelength in the spectrum that is
        dominated by scattering.

        Parameters
        ----------
        wavelength : astropy.units.quantity.Quantity
            The wavelengths on which we want the spectrum.
        '''


        # figure out the airmass along this line of sight
        # airmass = 1/np.cos(self.zenith_angle)
        effective_airmass = np.minimum(self.atmosphere.fortney_factor()/2.0,
                                       self.airmass)


        # calculate the optical depths at this angle
        tau_scatter = self.atmosphere.tau_zenith_scatter*effective_airmass
        tau_absorb = self.atmosphere.tau_zenith_absorb*effective_airmass

        # estimate the single
        albedo = tau_scatter/(tau_absorb + tau_scatter)

        # calculate the mean intensity field
        mean_intensity = self.sunset.mean_intensity()

        # calculate the intensity of the illuminated sky (away from the disk)
        thermal_intensity = (1 - albedo)*self.B(self.atmosphere.default_wavelengths)*(1 - np.exp(-tau_absorb))
        scattering_intensity = albedo * mean_intensity*(1 - np.exp(-tau_scatter))
        sky_intensity = thermal_intensity + scattering_intensity

        # bin this intensity onto the desired wavelengths
        w = self.wavelength(wavelength)
        unit = sky_intensity.unit
        neww, newi = bintogrid(self.atmosphere.default_wavelengths, sky_intensity.value,
                         newx=w.to('nm').value,
                         drop_nans=False)

        # make sure the wavelengths match up
        assert(np.all(neww == w.to('nm').value))


        # return this sky intensity
        return newi*unit



'''
A few shortcuts for dealing with units.
'''
import astropy.units as u

# define some useful units
spectral_luminosity_unit = u.W/u.nm
spectral_flux_unit = u.W/u.nm/u.m**2
spectral_intensity_unit = u.W/u.nm/u.m**2/u.sr

luminosity_unit = u.W
flux_unit = u.W/u.m**2
intensity_unit = u.W/u.m**2/u.sr

unit2name = {spectral_flux_unit:'Flux',
             spectral_luminosity_unit:'Luminosity',
             flux_unit:'Flux',
             luminosity_unit:'Luminosity',
             intensity_unit:'Intensity',
             spectral_intensity_unit:'Intensity'}
#name2unit = {v.si:k for k, v in unit2name.items()}

def determine_quantity(unit):
    '''
    Determine the name of a particular unit.

    Parameters
    ----------
    unit : astropy.units.core.Unit, astropy.units.core.CompositeUnit

    Returns
    -------
    '''

    # try all the listed units
    for k in unit2name:
        if unit.is_equivalent(k):
            return unit2name[k]

    # if none match, return "?"
    return '?'
